# -*- coding: utf-8 -*-

from .utils import safe_func
from events import Events


class RoutesMixin(object):
    """
    专门做路由管理
    """

    rule_map = None
    events = None

    def __init__(self):
        self.rule_map = dict()
        self.events = Events()

    def add_route_rule(self, cmd=None, view_func=None, **options):
        assert view_func is not None, 'expected view func if cmd is not provided.'

        cmd = cmd if cmd is not None else view_func.__name__

        if cmd in self.rule_map and view_func != self.rule_map[cmd]:
            raise Exception, 'duplicate view_func for cmd: %(cmd)s, old_view_func:%(old_view_func)s, new_view_func: %(new_view_func)s' % dict(
                cmd=cmd,
                old_view_func=self.rule_map[cmd],
                new_view_func=view_func,
            )

        self.rule_map[cmd] = view_func

    def route(self, cmd=None, **options):
        def decorator(f):
            self.add_route_rule(cmd, f, **options)
            return f
        return decorator

    def get_route_view_func(self, cmd):
        return self.rule_map.get(cmd)


class AppCallBacksMixin(RoutesMixin):

    def create_conn(self, f):
        """
        连接建立成功后
        f(conn)
        """
        self.events.create_conn += safe_func(f)
        return f

    def before_first_request(self, f):
        """
        第一个请求，请求解析为json成功后
        f(request)
        """
        self.events.before_first_request += safe_func(f)
        return f

    def before_request(self, f):
        """
        请求解析为json成功后
        f(request)
        """
        self.events.before_request += safe_func(f)
        return f

    def after_request(self, f):
        """
        执行完route对应的view_func后
        f(request, exc)
        """
        self.events.after_request += safe_func(f)
        return f

    def before_response(self, f):
        """
        在 stream.write 之前，传入encode之后的data
        f(conn, response)
        """
        self.events.before_response += safe_func(f)
        return f

    def after_response(self, f):
        """
        在 stream.write 之后，传入encode之后的data
        f(conn, response, result)
        """
        self.events.after_response += safe_func(f)
        return f

    def close_conn(self, f):
        """
        连接close之后
        f(conn)
        """
        self.events.close_conn += safe_func(f)
        return f


class BlueprintCallBacksMixin(RoutesMixin):

    def before_request(self, f):
        """
        请求解析为json成功后
        f(request)
        """
        self.events.before_request += safe_func(f)
        return f

    def after_request(self, f):
        """
        执行完route对应的view_func后
        f(request)
        """
        self.events.after_request += safe_func(f)
        return f

    def create_app_conn(self, f):
        """
        连接建立成功后
        f(conn)
        """
        self.events.create_app_conn += safe_func(f)
        return f

    def before_app_first_request(self, f):
        """
        第一次请求，请求解析为json成功后
        f(request)
        """
        self.events.before_app_first_request += safe_func(f)
        return f

    def before_app_request(self, f):
        """
        请求解析为json成功后
        f(request)
        """
        self.events.before_app_request += safe_func(f)
        return f

    def after_app_request(self, f):
        """
        执行完route对应的view_func后
        f(request)
        """
        self.events.after_app_request += safe_func(f)
        return f

    def before_app_response(self, f):
        """
        在 stream.write 之前，传入encode之后的data
        f(conn, response)
        """
        self.events.before_app_response += safe_func(f)
        return f

    def after_app_response(self, f):
        """
        在 stream.write 之后，传入encode之后的data
        f(conn, response, result)
        """
        self.events.after_app_response += safe_func(f)
        return f

    def close_app_conn(self, f):
        """
        连接close之后
        f(conn)
        """
        self.events.close_app_conn += safe_func(f)
        return f



# -*- coding: utf-8 -*-

from SocketServer import ThreadingTCPServer, StreamRequestHandler
import threading
import functools
from threading import Lock
from netkit.stream import Stream

from .connection import Connection
from .request import Request
from .haven import Haven
from .blueprint import Blueprint
from .utils import safe_call
from . import constants


class THaven(Haven):

    backlog = constants.SERVER_BACKLOG
    server = None
    got_first_request_lock = None

    def __init__(self, box_class, server_class=None, connection_class=None, request_class=None, stream_class=None):
        super(THaven, self).__init__()
        self.box_class = box_class
        self.server_class = server_class or ThreadingTCPServer
        self.connection_class = connection_class or Connection
        self.request_class = request_class or Request
        self.stream_class = stream_class or Stream
        self.got_first_request_lock = Lock()

    def repeat_timer(self, interval):
        def inner_repeat_timer(func):
            later = TTimer()
            self.events.repeat_timer += functools.partial(later.set, interval, func, True)
            return func

        return inner_repeat_timer

    def acquire_got_first_request(self):
        self.got_first_request_lock.acquire()

    def release_got_first_request(self):
        self.got_first_request_lock.release()

    def _prepare_server(self, host, port):
        class RequestHandler(StreamRequestHandler):
            def handle(sub_self):
                self.connection_class(
                    self, self.box_class, self.request_class, self.stream_class(sub_self.connection), sub_self.client_address
                ).handle()

        self.server = self.server_class((host, port), RequestHandler, bind_and_activate=False)
        self.server.request_queue_size = self.backlog
        # 主线程退出时，所有子线程结束
        self.server.daemon_threads = True
        # 必须在server_bind之前
        self.server.allow_reuse_address = True

        self.server.server_bind()
        self.server.server_activate()

    def _serve_forever(self):
        self.server.serve_forever()


class TBlueprint(Blueprint):

    def repeat_app_timer(self, interval):
        def inner_repeat_timer(func):
            later = TTimer()
            self.events.repeat_app_timer += functools.partial(later.set, interval, func, True)
            return func

        return inner_repeat_timer


class TTimer(object):

    timer = None

    def set(self, interval, callback, repeat=False, force=True):
        """
        添加timer
        """
        if self.timer:
            if force:
                # 如果已经存在，那么先要把现在的清空
                self.clear()
            else:
                # 已经存在的话，就返回了
                return

        def callback_wrapper():

            # 必须要确定，这次调用就是这个timer引起的
            if self.timer == timer:
                self.timer = None
                result = safe_call(callback)
                if repeat and not self.timer:
                    # 之所以还要判断timer，是因为callback中可能设置了新的回调
                    self.set(interval, callback, repeat, True)
                return result

        self.timer = timer = threading.Timer(interval, callback_wrapper)
        # 跟着主线程结束。默认值是和创建的线程一致
        timer.daemon = True
        timer.start()

    def clear(self):
        """
        直接把现在的清空
        """
        if not self.timer:
            return

        try:
            self.timer.cancel()
        except:
            pass
        self.timer = None



# -*- coding: utf-8 -*-

"""
websocket协议支持
"""

import re
import gevent.wsgi
from geventwebsocket.handler import WebSocketHandler
from netkit.stream import Stream
from haven import logger
from haven import GHaven


class WSStream(Stream):
    def read_from_fd(self):
        try:
            chunk = self.sock.receive()
        except:
            logger.error('exc occur.', exc_info=True)
            # 其他都直接关闭
            self.close()
            return None

        if not chunk:
            self.close()
            return None

        # 如果二进制传输，获取的将会是bytearray，会导致box.body是bytearray格式，从而导致protobuf parse报错
        return str(chunk)

    def write_to_fd(self, data):
        try:
            return self.sock.send(data)
        except:
            logger.error('exc occur. data: %r', data, exc_info=True)
            return None

    def close_fd(self):
        try:
            self.sock.close()
        except:
            logger.error('exc occur.', exc_info=True)
        finally:
            self.sock = None

    def shutdown_fd(self, how=2):
        try:
            self.sock.stream.handler.socket.shutdown(how)
        except:
            logger.error('exc occur.', exc_info=True)


class WSGHaven(GHaven):

    def __init__(self, box_class, path_pattern, merge_wsgi_app=None, **kwargs):
        # 去掉*args，即可避免stream_class在*args的情况
        if not kwargs.get('stream_class'):
            kwargs['stream_class'] = WSStream
        super(WSGHaven, self).__init__(box_class, **kwargs)

        self.path_pattern = path_pattern
        self.merge_wsgi_app = merge_wsgi_app

    def wsgi_app(self, environ, start_response):
        """
        将原始的app，包装为包括websocket的app
        """

        path = environ['PATH_INFO']

        if re.match(self.path_pattern, path):
            ws = environ['wsgi.websocket']
            address = (environ.get('REMOTE_ADDR'), environ.get('REMOTE_PORT'))
            self.handle_stream(ws, address)
        else:
            if self.merge_wsgi_app:
                return self.merge_wsgi_app(environ, start_response)
            else:
                start_response("400 Bad Request", [])
                return ['400 Bad Request']

    def __call__(self, environ, start_response):
        return self.wsgi_app(environ, start_response)

    def _prepare_server(self, host, port):
        self.server = gevent.wsgi.WSGIServer((host, port), self.wsgi_app,
                                             backlog=self.backlog, handler_class=WebSocketHandler)
        self.server.start()

    def _serve_forever(self):
        self.server.serve_forever()


# -*- coding: utf-8 -*-

from .callbacks_mixin import BlueprintCallBacksMixin


class Blueprint(BlueprintCallBacksMixin):

    name = None
    app = None

    def __init__(self, name=None):
        super(Blueprint, self).__init__()
        self.name = name

    def register2app(self, app):
        """
        注册到app上
        """
        self.app = app
        # 注册上
        self.app.blueprints.append(self)

    def repeat_app_timer(self, interval):
        raise NotImplementedError



# -*- coding: utf-8 -*-

from multiprocessing import Process
import time
import signal
from .callbacks_mixin import AppCallBacksMixin
from . import autoreload
from .log import logger
from . import constants


class Haven(AppCallBacksMixin):
    debug = False
    got_first_request = False
    blueprints = None

    def __init__(self):
        super(Haven, self).__init__()
        self.blueprints = list()

    def register_blueprint(self, blueprint):
        blueprint.register2app(self)

    def run(self, host=None, port=None, debug=None, use_reloader=None, workers=None, handle_signals=None):
        if host is None:
            host = constants.SERVER_HOST
        if port is None:
            port = constants.SERVER_PORT
        if debug is not None:
            self.debug = debug

        use_reloader = use_reloader if use_reloader is not None else self.debug
        handle_signals = handle_signals if handle_signals is not None else not use_reloader

        def run_wrapper():
            logger.info('Running server on %s:%s, debug: %s, use_reloader: %s',
                        host, port, self.debug, use_reloader)

            self._before_run()

            self._prepare_server(host, port)
            if workers is not None:
                if handle_signals:
                    # 因为只能在主线程里面设置signals
                    self._handle_parent_proc_signals()

                self._fork_workers(workers)
            else:
                self._try_serve_forever(True)

        if use_reloader:
            autoreload.main(run_wrapper)
        else:
            run_wrapper()

    def acquire_got_first_request(self):
        pass

    def release_got_first_request(self):
        pass

    def repeat_timer(self, interval):
        raise NotImplementedError

    def _before_run(self):
        self._start_repeat_timers()

    def _try_serve_forever(self, main_process):
        if not main_process:
            self._handle_child_proc_signals()

        try:
            self._serve_forever()
        except KeyboardInterrupt:
            pass
        except:
            logger.error('exc occur.', exc_info=True)

    def _fork_workers(self, workers):
        def start_worker_process():
            inner_p = Process(target=self._try_serve_forever, args=(False,))
            # 当前进程daemon默认是False，改成True将启动不了子进程
            # 但是子进程要设置daemon为True，这样父进程退出，子进程会被强制关闭
            inner_p.daemon = True
            inner_p.start()
            return inner_p

        p_list = []

        for it in xrange(0, workers):
            p = start_worker_process()
            p_list.append(p)

        while 1:
            for idx, p in enumerate(p_list):
                if not p.is_alive():
                    old_pid = p.pid
                    p = start_worker_process()
                    p_list[idx] = p

                    logger.error('process[%s] is dead. start new process[%s].', old_pid, p.pid)

            try:
                time.sleep(1)
            except KeyboardInterrupt:
                break
            except:
                logger.error('exc occur.', exc_info=True)
                break

    def _handle_parent_proc_signals(self):
        # 修改SIGTERM，否则父进程被term，子进程不会自动退出；明明子进程都设置为daemon了的
        signal.signal(signal.SIGTERM, signal.default_int_handler)
        # 即使对于SIGINT，SIG_DFL和default_int_handler也是不一样的，要是想要抛出KeyboardInterrupt，应该用default_int_handler
        signal.signal(signal.SIGINT, signal.default_int_handler)

    def _handle_child_proc_signals(self):
        signal.signal(signal.SIGTERM, signal.SIG_DFL)
        signal.signal(signal.SIGINT, signal.SIG_IGN)

    def _start_repeat_timers(self):
        self.events.repeat_timer()
        for bp in self.blueprints:
            bp.events.repeat_app_timer()

    def _prepare_server(self, host, port):
        raise NotImplementedError

    def _serve_forever(self):
        raise NotImplementedError
